/*
**  coords.c
**
**  Shape From Motion utility.
**
**  
**  Shape From Motion is used to derive 3D information about an object by
**  analyzing images of its rotation around a vertical axis.
**
**  This utility takes as input the .TGA frames specified in the command line
**  and provides as output a .RAW file containing the 3D triangular surfaces
**  it derives from the frames.  This file can be imported directly into the
**  Persistence of Vision Raytracer or can be converted to .DXF format for use
**  in popular CAD applications using pre-existing conversion utilities.
**
**  This software may be redistributed freely as long as any modifications of
**  the code are well documented and include the following copyright.
**
**  Copyright (C) 1995, Benjamin Nason Lipchak
*/

#include "coords.h"

/******************************************************************************/

/*
**  This module converts cylindrical coordinates to rectangular coordinates.
*/

int coords( char *out, double *radii, int quantum, int top, int bottom, 
            int height, int p, int t_p, int r, int a, int v )
{
   int ty, ta;
   float angle;
   float *rectangular_x, *rectangular_y, *rectangular_z;

/*
**  Allocate space for rectangular coordinates.
*/

   if ( ( rectangular_x = (float *)malloc( sizeof(float) * ( (bottom - top + 1) 
                                           * quantum + 2 ) ) ) == NULL )
   {
      printf( "Error allocating memory!\n" );
      return -1;
   }

   if ( ( rectangular_y = (float *)malloc( sizeof(float) * ( (bottom - top + 1)
                                           * quantum + 2 ) ) ) == NULL )
   {
      printf( "Error allocating memory!\n" );
      free( rectangular_x );
      return -1;
   }
    
   if ( ( rectangular_z = (float *)malloc( sizeof(float) * ( (bottom - top + 1)
                                           * quantum + 2 ) ) ) == NULL )
   {
      printf( "Error allocating memory!\n" );
      free( rectangular_x );
      free( rectangular_y );
      return -1;
   }
    
/*
**  Convert to rectangular coordinates.
*/

   rectangular_x[0] = 0.0;
   rectangular_y[0] = 0.0;
   rectangular_z[0] = 0.5 * (float)(bottom-top) * 
                         (float)height / (float)(quantum+1);

   rectangular_x[1] = 0.0;
   rectangular_y[1] = 0.0;
   rectangular_z[1] = -rectangular_z[0];

   for ( ty = top; ty <= bottom; ty++ )
      for ( ta = 0; ta < quantum; ta++ )
      {
         angle = (float)ta * (2.0 * PI / (float)quantum);
       
         if ( a == Y )
         {
            rectangular_x[(ty-top) * quantum + ta + 2] = 
               radii[ty * quantum + ta] * cos( angle );

            rectangular_y[(ty-top) * quantum + ta + 2] = 
               radii[ty * quantum + ta] * sin( angle );

            rectangular_z[(ty-top) * quantum + ta + 2] = rectangular_z[0] - 
               ( (float)(ty-top) * (float)height / (float)(quantum + 1) ); 
         }  else
         {     
            rectangular_y[(ty-top) * quantum + ta + 2] = 
               radii[ty * quantum + ta] * cos( angle );

            rectangular_z[(ty-top) * quantum + ta + 2] = 
               radii[ty * quantum + ta] * sin( angle );

            rectangular_x[(ty-top) * quantum + ta + 2] = rectangular_z[0] - 
               ( (float)(ty-top) * (float)height / (float)(quantum + 1) ); 
         }         
         
         if ( abs( rectangular_x[(ty-top) * quantum + ta + 2] )
              < .0001 )
            rectangular_x[(ty-top) * quantum + ta + 2] = 0.0;
         
         if ( abs( rectangular_y[(ty-top) * quantum + ta + 2] )
              < .0001 )
            rectangular_y[(ty-top) * quantum + ta + 2] = 0.0;
         
         if ( abs( rectangular_z[(ty-top) * quantum + ta + 2] )
              < .0001 )
            rectangular_z[(ty-top) * quantum + ta + 2] = 0.0;
      }

/*
**  Group coordinates into triangular surfaces.
*/

   if ( group( rectangular_x, rectangular_y, rectangular_z, (bottom - top + 1),
               quantum, p, t_p, r, v, out ) )
   {
      printf( "An error occurred while grouping!\n" );
      free( rectangular_x );
      free( rectangular_y );
      free( rectangular_z );
      return -1;
   }

/*
**  Free dynamically allocated memory.
*/

   free( rectangular_x );
   free( rectangular_y );
   free( rectangular_z );   
   
/*
**  Return gracefully.
*/
   
   return 0;
}