/*
**  corr.c
**
**  Shape From Motion utility.
**
**  
**  Shape From Motion is used to derive 3D information about an object by
**  analyzing images of its rotation around a vertical axis.
**
**  This utility takes as input the .TGA frames specified in the command line
**  and provides as output a .RAW file containing the 3D triangular surfaces
**  it derives from the frames.  This file can be imported directly into the
**  Persistence of Vision Raytracer or can be converted to .DXF format for use
**  in popular CAD applications using pre-existing conversion utilities.
**
**  This software may be redistributed freely as long as any modifications of
**  the code are well documented and include the following copyright.
**
**  Copyright (C) 1995, Benjamin Nason Lipchak
*/

#include "corr.h"

/******************************************************************************/

/*
**  Correspondence detection module
**
**  Takes images two at a time and matches points.
**  The result is a bunch of radii.
*/

int corr( unsigned char **images, char *out, int nf, int c, int a, int d, 
          int p, int t_p, int r, int v, int ms, int mo, double b, 
          int width, int height, int q_i, int q_s, int q_e )
{
   int x, y, i, j, flag, first, second;
   double best_match, temp_match, inc = (2.0 * PI) / (double)nf;
   long int total_radii = 0;
   double *radii;
   unsigned char red, green, blue, gray, test_val;

/*
**  Begin correspondence matching...
*/

   if ( ( radii = (double *)malloc( sizeof( double ) * 30000 ) ) == NULL )
   {
      printf( "Unable to allocate memory for radii!\n" );
      return -1;
   }
   
   for ( j = 0; j < nf; j++ )
   {

/*
**  Pick frames depending on direction of rotation.
*/

      if ( d == CCW )
      {
         first = j;
         second = (j + 1) % nf;
      }  else
      {
         first = j;
         second = (j + nf - 1) % nf;
      }
         
      if ( j == 0 && v == ON )
         printf( "\nComparing frames " );
      else if ( v == ON )
         printf( "\b\b\b\b\b\b\b\b\b\b\b\b\b" );
      if ( v == ON )
      {
         printf( "%02d and %02d... ", first, second );
         fflush( stdout );
      }
         
      for ( y = (ms - 1) / 2; y < height - ((ms - 1) / 2); y++ )
      {
         for ( x = (ms - 1) / 2; x < width - ((ms - 1) / 2); x++ )

/*
**  Check to make sure each matrix actually moves and has an edge in it.
*/

            if ( match( images[first], images[second], ms, x, y, 0, width ) 
                 != 1.0 && pcntedge( images[first], ms, x, y, width ) > .2 )
            {
               best_match = 0.0; flag = 0;

/*
**  Find best match.
*/

               for ( i = 1; i < mo && x + i + ((ms - 1) / 2) < width; i++ )
               {
                  temp_match = match( images[first], images[second], ms, 
                                      x, y, i, width );
                  if ( temp_match > best_match )
                  {
                     best_match = temp_match;
                     flag = i;
                  }   
               }   
/*
**  Make sure match is a worthy one.
*/
               
               if ( best_match > b )
               {
                  double rad, angle, orig, off;
                  
                  orig = (double)( x - c );
                  off = (double)( x - c + flag );
      
                  if ( orig == 0 )
                  {
                     angle = 1.5 * PI;
                     rad = off / cos( angle + inc );
                  }
                  else
                  {
                     angle = 2 * PI + atan2( ( off / orig ) - cos( inc ), 
                                             -sin( inc ) ); 
                     if ( orig >= 0 && off >= 0 )
                        angle -= PI;
      
                     rad = orig / cos( angle );
                  }
                  if ( rad > 0.0 )
                  {
                     if ( d == CCW )
                        radii[total_radii*3] = 
                              fmod( ( angle + (2.0 * PI) - (double)j * inc ), 
                                    (2.0 * PI) );
                     else
                        radii[total_radii*3] = 
                              fmod( ( angle + (2.0 * PI) + (double)j * inc ), 
                                    (2.0 * PI) );
                     radii[total_radii*3 + 1] = (double)y;
                     radii[total_radii*3 + 2] = rad;
                     
                     total_radii++;
                     
/*
**  Reallocate memory for radii as needed.
*/
                     
                     if ( ( total_radii % 10000 ) == 0 )
                        if ( ( radii = (double *)realloc( radii, sizeof( double ) * 3 *
                                       (total_radii + 10000 ) ) ) == NULL )
                        {
                           printf( "Unable to allocate memory for radii!\n" );
                           free( radii );
                           return -1;
                        }
                  }
               }  
            }
      }
   }

/*
**  Free edge image memory.
*/

   for ( j = 0; j < nf; j++ )
      free( images[j] );

/*
**  Send radii to be quantized and merged.
*/

   if ( merge( radii, total_radii, out, height, p, t_p, r, v, 
               a, c, q_i, q_s, q_e ) )
   {
      printf( "An error occured in the output module!\n" );
      free( radii );
      return -1;
   }
  
/*
**  Free dynamically allocated memory.
*/

   free( radii );
         
/*
**  Exit gracefully
*/      
         
   return 0;
}

/*****************************************************************************/

double match( unsigned char *i1, unsigned char *i2, 
              int m_s, int x, int y, int o, int width )
{
   int i, j;
   double total = 0.0;

   for ( j = -(m_s - 1) / 2; j < (m_s + 1) / 2; j++ )
      for ( i = -(m_s - 1) / 2; i < (m_s + 1) / 2; i++ )
         if ( i1[(y+j)*width + x+i] == i2[(y+j)*width + x+i+o] )
            total += 1.0;
   
   total /= (double)(m_s * m_s);
   
   return total;   
}

/*****************************************************************************/

double pcntedge( unsigned char *image, int m_s, int x, int y, int width ) 
{
   int i, j;
   double total = 0.0;

   for ( j = -(m_s - 1) / 2; j < (m_s + 1) / 2; j++ )
      for ( i = -(m_s - 1) / 2; i < (m_s + 1) / 2; i++ )
         if ( image[(y + j)*width + x + i] == 0 )
            total += 1.0;
   
   total /= (double)(m_s * m_s);
   
   return total;   
}