/*
**  defaults.c
**
**  Shape From Motion utility.
**
**  
**  Shape From Motion is used to derive 3D information about an object by
**  analyzing images of its rotation around a vertical axis.
**
**  This utility takes as input the .TGA frames specified in the command line
**  and provides as output a .POV file containing the 3D triangular surfaces
**  it derives from the frames.  This file can be imported directly into the
**  Persistence of Vision Raytracer or can be converted to .DXF format for use
**  in popular CAD applications using pre-existing conversion utilities.
**
**  This software may be redistributed freely as long as any modifications of
**  the code are well documented and include the following copyright.
**
**  Copyright (C) 1995, Benjamin Nason Lipchak
*/

#include "defaults.h"

/******************************************************************************/

/*
**  This module finalizes the parameters for edge detection and correspondence
**  by checking that the values make sense and filling the values that require
**  image inspection.
*/

int defaults( char *base_string, char *output, int nf, int c, int a, int d, 
              int p, int t_p, int r, int v, int q_i, int q_s, int q_e, 
              double s, int ms, double b )
{
   int i, flag, height, width, new_width = 0, new_height = 0, name_length;
   char *filename;
   FILE *in;
   
/*
**  Load image headers to find height, width, num_frames.
*/

   name_length = strlen( base_string ) + 8;
   if ( ( filename = (char *)malloc( name_length ) ) == NULL )
   {
      printf( "Error allocating memory!\n" );
      return -1;
   }

   for ( i = 0, flag = 0; flag == 0; i++ )
   {
      sprintf( filename, "%s%03d.tga", base_string, i );
      
      if ( ( in = fopen( filename, "r" ) ) == NULL )
         flag = 1;
      else
      {
         fseek( in, 12, SEEK_SET );
         fread( &new_width, 2, 1, in );
         fread( &new_height, 2, 1, in );

         if ( i == 0 )
         {
            width = new_width;
            height = new_height;
         }
                    
         if ( new_width != width || new_height != height )
            flag = -1;
         
         fclose( in );
      }          
   }

   if ( flag == -1 )
   {
      printf( "\nAll images must have same dimensions!\n" );
      printf( "%s%03d.tga had different dimensions.\n\n", base_string, i - 1 );
      free( filename );
      return -1;
   }  else
   {
      if ( nf == LATER )
         nf = i - 1;
      
      if ( nf < 4 )
      {
         printf( "\nThere must be at least four frames!\n\n" );
         free( filename );
         return -1;
      } 
   }

/*
**  Check that other parameters make sense.
*/

   if ( s < 0.5 || s > 4.0 )
   {
      if ( v == ON )
         printf( "\nRejecting sigma value of %g!\n\n", s );
      s = 1.75;
   }

   if ( ( ms < 3 || ms > 51 ) && ms != LATER )
   {
      if ( v == ON )
         printf( "\nRejecting matrix size of %d!\n\n", ms );
      ms = LATER;
   }

   if ( a == Y )
      if ( ( c < 0 || c > width ) && c != LATER )
      {
         if ( v == ON )
            printf( "\nRejecting center of %d!\n\n", c );
         c = LATER;
      }
   else
      if ( ( c < 0 || c > height ) && c != LATER )
      {
         if ( v == ON )
            printf( "\nRejecting center of %d!\n\n", c );
         c = LATER;
      }

   if ( q_s < 8 || q_s > 360 )
   {
      if ( v == ON )
         printf( "\nRejecting quantum start of %d!\n\n", q_s );
      q_s = 72;
   }

   if ( q_e < 8 || q_e > 360 )
   {
      if ( v == ON )
         printf( "\nRejecting quantum end of %d!\n\n", q_e );
      q_e = 24;
   }

   if ( q_i < 1 || q_i > q_s )
   {
      if ( v == ON )
         printf( "\nRejecting quantum increment of %d!\n\n", q_i );
      q_i = 12;
   }

/*
**  Matrix size is based on image size in direction of axis.
*/
     
   if ( ms == LATER )
   {
      if ( a == Y )
         ms = (int)rint( (double)height / 32.0 );
      else
         ms = (int)rint( (double)width / 32.0 );
      
      if ( ms % 2 == 0 )
         ms--;
   }
   
/*
**  If verbose, show updated parameters.
*/

   if ( v == ON )
   {
      printf( "\nNew options:     -o%s -n", output );
      
      if ( nf == LATER )
         printf( "? -c" );
      else
         printf( "%d -c", nf );

      if ( c == LATER )
         printf( "? -a" );
      else
         printf( "%d -a", c );
         
      if ( a == Y )
         printf( "Y -d" );
      else
         printf( "X -d" );

      if ( d == CW )
         printf( "CW -r" );
      else
         printf( "CCW -r" );

      if ( r == ON )
         printf( "ON -p" );
      else
         printf( "OFF -p" );

      if ( p == ON )
         printf( "ON -t" );
      else
         printf( "OFF -t" );
         
      if ( t_p == ON )
         printf( "ON -v" );
      else
         printf( "OFF -v" );

      printf( "ON\n                 -s%g -m", s ); 

      printf( "%d -b%g -i%d -j%d -k%d\n\n", ms, b, q_i, q_s, q_e );
   }

/*
**  Send parameters to begin edge detection.
*/

   if ( edges( base_string, output, nf, c, a, d, p, t_p, r, v, 
               s, ms, b, width, height, q_i, q_s, q_e ) )
   {
      printf( "Error in edge detection module!\n\n" );
      free( filename );
      return -1;
   }
      
/*
**  Free dynamically allocated memory.
*/

   free( filename );

/*
**  Exit gracefully.
*/
   return 0;
}