/*
**  edges.c
**
**  Shape From Motion utility.
**
**  
**  Shape From Motion is used to derive 3D information about an object by
**  analyzing images of its rotation around a vertical axis.
**
**  This utility takes as input the .TGA frames specified in the command line
**  and provides as output a .POV file containing the 3D triangular surfaces
**  it derives from the frames.  This file can be imported directly into the
**  Persistence of Vision Raytracer or can be converted to .DXF format for use
**  in popular CAD applications using pre-existing conversion utilities.
**
**  This software may be redistributed freely as long as any modifications of
**  the code are well documented and include the following copyright.
**
**  Copyright (C) 1995, Benjamin Nason Lipchak
*/

#include "edges.h"

/******************************************************************************/

/*
**  This module stores all edge images in memory, finds the center of
**  rotation, and then calculates the maximum_offset.
*/

int edges( char *bs, char *out, int nf, int c, int a, int d, 
           int p, int t_p, int r, int v, double s, int ms, double b, 
           int width, int height, int q_i, int q_s, int q_e )
{           

   unsigned char **images;
   int kernel_size, i, j, new_width, new_height, base_length;
   int left, right, max_right = 0, center_total = 0, total_points, x, y, mo;
   char *filename;
   double *kernel, rad, t = 0.001;
   
/*
**  Allocating space for image pointers.
*/

   if ( ( images = (unsigned char **)malloc( sizeof( unsigned char *) * nf ) ) == NULL )
   {
      printf( "Unable to allocate memory for image pointers!\n" );
      return -1;
   }   

/*
**  Calculate minimum kernel size required for laplacian of gaussian matrix.
*/
   
   kernel_size = get_kernel_size( s, t );

   if ( (kernel = (double *)malloc( sizeof(double)*kernel_size*kernel_size ) )
        == NULL )
   {
      printf( "Unable to allocate memory for kernel!\n" );
      free( images );
      return -1;
   }
   
/*
**  Fill kernel with Laplacian of the Gaussian.
*/

   for ( i = 0; i < kernel_size; i++ )
      for ( j = 0; j < kernel_size; j++ )
      {
         rad = sqrt( ( ( i - ( ( kernel_size - 1 ) / 2 ) ) *
                     ( i - ( ( kernel_size - 1 ) / 2 ) ) ) +
                   ( ( j - ( ( kernel_size - 1 ) / 2 ) ) *
                     ( j - ( ( kernel_size - 1 ) / 2 ) ) ) );
                     
         kernel[j*kernel_size + i] = ( -1.0 / (PI*s*s*s*s) ) *
                                     ( 1.0 - ( (rad*rad) / (2.0*s*s) ) ) *
                                     exp( -(rad*rad) / (2.0*s*s) );
      }

/*
**  Allocate memory for all edge images.
*/

   if ( a == Y )
   {
      new_width = width - (kernel_size - 1);
      new_height = height - (kernel_size - 1);
   }  else
   {
      new_width = height - (kernel_size - 1);
      new_height = width - (kernel_size - 1);
   }
   
   for ( i = 0; i < nf; i++ )
      if ( ( images[i] = (unsigned char *)malloc( new_width * new_height ) ) == NULL )
      {
         printf( "Unable to allocate memory for image %d!\n\n", i + 1 );
         for ( j = 0; j < i; j++ )
            free( images[j] );
         free( kernel );
         free( images );
         return -1;
      }
      
/*
**  Send each image file for edge detection.
*/   
   
   base_length = strlen( bs );
      
   if ( ( filename = (char *)malloc( base_length + 8 ) ) == NULL )
   {
      printf( "Unable to allocate memory for file name!\n" );
      for ( j = 0; j < i; j++ )
         free( images[j] );
      free( filename );
      free( kernel );
      free( images );
      return -1;
   }      
      
   for ( i = 0; i < nf; i++ )
   {
      sprintf( filename, "%s%03d.tga", bs, i );
      
      if ( i == 0 && v == ON )
         printf( "\nOpening " );
      else if ( v == ON )
      {
         printf( "\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b" );
         for ( j = 0; j < strlen( filename ); j++ )
            printf( "\b" );
         fflush( stdout );
      }
      
      if ( edge_detect( s, t, images[i], filename, kernel, kernel_size, a, v ) )
      {
         printf( "An error occurred while edge detecting frame %d!\n", i + 1 );
         for ( j = 0; j < i; j++ )
            free( images[j] );
         free( filename );
         free( kernel );
         free( images );
         return -1;
      }
   }      

/*
**  Find the extreme left and right of object in all frames.  The extreme
**  left in one frame corresponds to the extreme right in the frame at
**  (frame + (nf/2)) mod nf.
*/      
      
   for ( i = 0; i < nf; i++ )
   {
      for ( x = 0, left = -1; x < new_width && left == -1; x++ )
      {
         for ( y = 0, total_points = 0; y < new_height; y++ )
            if ( images[i][y*new_width + x] == 0 )
               total_points++;
               
         if ( (double)total_points > (double)height * .15 )
            left = x;       
      }

      for ( x = new_width, right = -1; x >= 0 && right == -1; x-- )
      {
         for ( y = 0, total_points = 0; y < new_height; y++ )
            if ( images[(i+(nf/2))%nf][y*new_width + x] == 0 )
               total_points++;
               
         if ( (double)total_points > (double)height * .15 )
            right = x;       
            
         if ( right > max_right )
            max_right = right;
      }
      
      center_total += ( left + ((right - left) / 2 ) );   
   }
   
   if ( c == LATER )
      c = center_total / nf;
   else
      c -= (kernel_size-1) / 2;
   
/*
**  Calculate the maximum offset based on the maximum estimated radius.
*/

   max_right -= c;
   
   mo = (int)ceil( (double)max_right * 2.0 * sin( PI / (double)nf ) );

/*
**  If verbose, show updated parameters.
*/

   if ( v == ON )
   {
      printf( "\n\n\nFinal options:   -o%s -n%d -c%d -a", 
              out, nf, c + ( (kernel_size-1) / 2 ) );
      
      if ( a == Y )
         printf( "Y -d" );
      else
         printf( "X -d" );

      if ( d == CW )
         printf( "CW -r" );
      else
         printf( "CCW -r" );

      if ( r == ON )
         printf( "ON -p" );
      else
         printf( "OFF -p" );

      if ( p == ON )
         printf( "ON -t" );
      else
         printf( "OFF -t" );
         
      if ( t_p == ON )
         printf( "ON -v" );
      else
         printf( "OFF -v" );

      printf( "ON\n                 -s%g -m%d -b%g -i%d -j%d -k%d\n\n", 
              s, ms, b, q_i, q_s, q_e ); 
   }

/*
**  Send final parameters to correspondence detection.
*/
   
   if ( corr( images, out, nf, c, a, d, p, t_p, r, v, ms, mo, b, 
              new_width, new_height, q_i, q_s, q_e ) )
   {
      printf( "An error occurred in the correspondence detection module!\n" );
      free( images );
      free( filename );
      free( kernel );
   }   
      
/*
**  Free dynamically allocated memory.
*/

   free( filename );
   free( kernel );
   free( images );

/*
**  Exit gracefully.
*/
 
   return 0;
}

/*****************************************************************************/

int edge_detect( double sigma, double tolerance, unsigned char *image_out, 
                 char *filename, double *kernel, int kernel_size, int a, int v )
{
   int x, y, i, j, num_pos, num_neg;
   int temp, new_width, new_height, width = 0, height = 0;
   unsigned char red, green, blue, gray, test_val;
   char edge_val;
   unsigned char *image_in, *image_reg;
   double value, r;
   FILE *in;

/*
**  Load image into memory.
*/

   if ( ( in = fopen( filename, "r" ) ) == NULL )
   {
      printf( "Unable to open input file %s!\n", filename );
      return -1;
   }  else                           
   {
      fseek( in, 2, SEEK_SET );
      fread( &test_val, 1, 1, in );
      if ( test_val != 2 )
      {
         printf( "%s not a recognized Targa file!\n", filename );
         fclose( in );
         return -1;
      } 
      if ( fseek( in, 12, SEEK_SET ) )
      {
         printf( "%s not a recognized Targa file!\n", filename );
         fclose( in );
         return -1;
      } 
      fread( &width, 2, 1, in );
      fread( &height, 2, 1, in );
   }

   if ( a == X )
   {
      temp = height;
      height = width;
      width = temp;
   }

   if ( v == ON )
   {
      printf( "%s (%03d x %03d) for edge detection... ", filename, width, height );
      fflush( stdout );
   }
   
   if ( ( image_in = (unsigned char *)malloc( height * width ) ) == NULL )
   {
      printf( "Unable to allocate memory for image!\n" );
      fclose( in );
      return -1;
   }

   fseek( in, 18, SEEK_SET );
   
   if ( a == Y )
      for ( y = 0; y < height; y++ )
         for ( x = 0; x < width; x++ )
         {
            fread( &red, 1, 1, in );
            fread( &green, 1, 1, in );
            fread( &blue, 1, 1, in );
            gray = (unsigned char)( (red + green + blue ) / 3 );
            image_in[y*width + x] = gray;
         }
   else
      for ( x = 0; x < width; x++ )
         for ( y = 0; y < height; y++ )
         {
            fread( &red, 1, 1, in );
            fread( &green, 1, 1, in );
            fread( &blue, 1, 1, in );
            gray = (unsigned char)( (red + green + blue ) / 3 );
            image_in[(height-y-1)*width + x] = gray;
         }

   fclose( in );
   
/*
**  Multiply matrices for each pixel.
*/

   new_width = width - (kernel_size - 1);
   new_height = height - (kernel_size - 1);

   if ( ( image_reg = (unsigned char *)malloc( new_height*new_width ) ) == NULL )
   {
      printf( "Unable to allocate memory for region image!\n" );
      free( kernel );
      free( image_in );
      return -1;
   }
   
   for ( y = (kernel_size - 1) / 2; y < height - ((kernel_size - 1) / 2); y++ )
      for ( x = (kernel_size - 1) / 2; x < width - ((kernel_size - 1) / 2); x++ )
      {
         edge_val = matrix_mult( kernel, image_in, x, y, kernel_size, width );
         if ( edge_val < 0 )
            gray = 0;
         else
            gray = 255;
         
         image_reg[(y-((kernel_size-1)/2))*new_width + 
                   x-((kernel_size-1)/2)] = gray;
      }   

/*
**  Make edge regions into edge contours.
*/

   for ( y = 0; y < new_height; y++ )
      for ( x = 0; x < new_width; x++ )
      {
         if ( x < 1 || y < 1 || x > new_width - 2 || y > new_height - 2 )
            gray = 255;
         else
         {
            num_pos = 0; num_neg = 0;
            for ( i = -1; i < 2; i++ )
               for ( j = -1; j < 2; j++ )
               {
                  if ( image_reg[(y+j)*new_width + x+i] == 255 )
                     num_pos++;
                  else if ( image_reg[(y+j)*new_width + x+i] == 0 )
                     num_neg++;
               }
            
            if ( num_pos >= 3 && num_neg >= 3 )
               gray = 0;
            else
               gray = 255;
         }   

         image_out[y*new_width + x] = gray;
      }            
   
/*
**  Free dynamically allocated memory.
*/

   free( image_in );
   free( image_reg );

/*
**  Graceful exit.
*/
   
   return 0;
}

/*****************************************************************************/

char matrix_mult( double *kernel, unsigned char *image, 
                  int x, int y, int k_s, int width )
{
   double total = 0.0;
   double image_val;
   int i, j;
    
   for ( j = 0; j < k_s; j++ )
      for ( i = 0; i < k_s; i++ )
      {
         image_val = (double)image[(y - ((k_s - 1) / 2) + j) * width + 
                                   (x - ((k_s - 1) / 2) + i)];
         
         total += image_val * kernel[j*k_s + i];
      }
   
   return (char)total;      
}

/*****************************************************************************/

/*
**  Algorithm made public to CS 549 by Norman Wittels
*/

/****** calculate the minimum kernel size for a given sigma ******/
int get_kernel_size(double sigma, double tolerance)
	{
	int r;
	double a,b;
	
	if (sigma<=0.0)
		{
		printf("get_kernel_size() was passed a non-positive value of sigma\n");
		return 0;
		}
	for (r=1; r<50; r++)
		{
		a = (double)r / sigma;
		a *= a;
		b = (1.0 + a) * exp(-a/2.0);
		if (b < tolerance) return (2*r+1);
		} /* end for(r) */
	printf("get_kernel_size() could not find a satisfactory kernel smaller than 101\n");
	return 0;
	} /* end get_kernel_size() */

              