/*
**  group.c
**
**  Shape From Motion utility.
**
**  
**  Shape From Motion is used to derive 3D information about an object by
**  analyzing images of its rotation around a vertical axis.
**
**  This utility takes as input the .TGA frames specified in the command line
**  and provides as output a .RAW file containing the 3D triangular surfaces
**  it derives from the frames.  This file can be imported directly into the
**  Persistence of Vision Raytracer or can be converted to .DXF format for use
**  in popular CAD applications using pre-existing conversion utilities.
**
**  This software may be redistributed freely as long as any modifications of
**  the code are well documented and include the following copyright.
**
**  Copyright (C) 1995, Benjamin Nason Lipchak
*/

#include "export.h"

/******************************************************************************/

/*
**  In order to use the triangular surfaces in a 3D model, they must be in
**  a format readable by 3D modeling software.  Such formats are .raw, .pov,
**  and .dxf.  Right now .raw and .pov will be supported.
*/

int export( float *t1, float *t2, float *t3, int num_t, 
            int p, int t_p, int r, int v, char *out )
{
   int name_length, i;
   char *raw_name, *pov_name;
   FILE *raw, *pov;
   char texture[14] = "PinkAlabaster";
   float *n1, *n2, *n3;

   name_length = strlen( out ) + 5;
   if ( ( raw_name = (char *)malloc( name_length ) ) == NULL )
   {
      printf( "Error allocating memory!\n" );
      return -1;
   }

   if ( ( pov_name = (char *)malloc( name_length ) ) == NULL )
   {
      printf( "Error allocating memory!\n" );
      free( raw_name );
      return -1;
   }
   
   sprintf( raw_name, "%s.raw", out );
   sprintf( pov_name, "%s.pov", out );
      
   if ( r == ON )
      if ( ( raw = fopen( raw_name, "w" ) ) == NULL )
      {
         printf( "Could not open file %s!\n", raw_name );
         free( raw_name );
         free( pov_name );
         return -1;
      }              

   if ( p == ON || t_p == ON )
   {
      if ( ( pov = fopen( pov_name, "w" ) ) == NULL )
      {
         printf( "Could not open file %s!\n", pov_name );
         free( raw_name );
         free( pov_name );
         return -1;
      }              

      fprintf( pov, "#include \"colors.inc\"\n" );
      fprintf( pov, "#include \"textures.inc\"\n" );
      fprintf( pov, "#include \"shapes.inc\"\n\n" );
   
      fprintf( pov, "camera {\n" );
      fprintf( pov, "\tlocation  <0.00, -500.00, 0.00>\n" );
      fprintf( pov, "\tdirection <0.00, 0.00, 1.45>\n" );
      fprintf( pov, "\tsky       <0.00, 0.00, 1.00>\n" );
      fprintf( pov, "\tup        <0.00, 0.00, 1.00>\n" );
      fprintf( pov, "\tright     <1.333, 0.00, 0.00>\n" );
      fprintf( pov, "\tlook_at   <0.00, 0.00, 0.00>\n}\n\n" );
   
      fprintf( pov, "light_source {\n" );
      fprintf( pov, "\t<0.00, -300.0, 0.00>\n" );
      fprintf( pov, "\tcolor rgb <1.00, 1.00, 1.00>\n}\n\n" );
   
      fprintf( pov, "object {\n\tunion {\n" );
   }
   
   for ( i = 0; i < num_t; i++ )
   {
      if ( r == ON )
         fprintf( raw, "%#.3f %#.3f %#.3f  %#.3f %#.3f %#.3f  %#.3f %#.3f %#.3f\n",
                  t1[i * 3], t1[i * 3 + 1], t1[i * 3 + 2],
                  t2[i * 3], t2[i * 3 + 1], t2[i * 3 + 2],
                  t3[i * 3], t3[i * 3 + 1], t3[i * 3 + 2] );
      
      if ( p == ON )
      {         
         n1 = findnorm( t1[i*3], t1[i*3 + 1], t1[i*3 + 2], t1, t2, t3, num_t );
         n2 = findnorm( t2[i*3], t2[i*3 + 1], t2[i*3 + 2], t1, t2, t3, num_t );
         n3 = findnorm( t3[i*3], t3[i*3 + 1], t3[i*3 + 2], t1, t2, t3, num_t );
         
         fprintf( pov, "\t\tsmooth_triangle {\n" );
         fprintf( pov, "\t\t\t<%#.3f, %#.3f, %#.3f> ", 
                  t1[i * 3], t1[i * 3 + 1], t1[i * 3 + 2] );
         fprintf( pov, "<%#.3f, %#.3f, %#.3f>\n", n1[0], n1[1], n1[2] );
         fprintf( pov, "\t\t\t<%#.3f, %#.3f, %#.3f> ", 
                  t2[i * 3], t2[i * 3 + 1], t2[i * 3 + 2] );
         fprintf( pov, "<%#.3f, %#.3f, %#.3f>\n", n2[0], n2[1], n2[2] );
         fprintf( pov, "\t\t\t<%#.3f, %#.3f, %#.3f> ", 
                  t3[i * 3], t3[i * 3 + 1], t3[i * 3 + 2] );
         fprintf( pov, "<%#.3f, %#.3f, %#.3f>\n", n3[0], n3[1], n3[2] );
         fprintf( pov, "\t\t}\n" );
      }  else if ( t_p == ON ) 
      {
         fprintf( pov, "\t\ttriangle {\n" );
         fprintf( pov, "\t\t\t< %#.3f, %#.3f, %#.3f >\n", 
                  t1[i * 3], t1[i * 3 + 1], t1[i * 3 + 2] );
         fprintf( pov, "\t\t\t< %#.3f, %#.3f, %#.3f >\n", 
                  t2[i * 3], t2[i * 3 + 1], t2[i * 3 + 2] );
         fprintf( pov, "\t\t\t< %#.3f, %#.3f, %#.3f >\n", 
                  t3[i * 3], t3[i * 3 + 1], t3[i * 3 + 2] );
         fprintf( pov, "\t\t}\n" );
      }
   }            
   
   if ( p == ON || t_p == ON )
   {
      fprintf( pov, "\t}\n\ttexture { %s }\n", texture );
      fprintf( pov, "\trotate <0.00, 0.00, 0.00>\n}\n" );
      fclose( pov );
   }
   
   if ( r == ON )
      fclose( raw ); 

/*
**  Free dynamically allocated memory.
*/

   free( raw_name );
   free( pov_name );
   
/*
**  Exit gracefully.
*/

   return 0;
}

/******************************************************************************/

/*
**  Finds all triangles that share the given vertex, then averages
**  the normals of each triangle to arrive at the normal for the point.
*/

float *findnorm( float x, float y, float z, 
                 float *t1, float *t2, float *t3, int nt )
{
   float *n;
   float *temp;
   int i, count = 0;
   
   n = (float *)malloc( 3 * sizeof( float ) );
   n[0] = 0.0;
   n[1] = 0.0;
   n[2] = 0.0; 

   temp = (float *)malloc( 3 * sizeof( float ) );
   
   for ( i = 0; i < nt; i++ ) 
      if ( ( t1[i*3] == x && t1[i*3+1] == y && t1[i*3+2] == z ) ||
           ( t2[i*3] == x && t2[i*3+1] == y && t2[i*3+2] == z ) ||
           ( t3[i*3] == x && t3[i*3+1] == y && t3[i*3+2] == z ) )
      {
         memcpy( temp, normal( &t1[i*3], &t2[i*3], &t3[i*3] ), 
                 sizeof( float ) * 3 );
         n[0] += temp[0];
         n[1] += temp[1];
         n[2] += temp[2];
         count++;
      }
      
   n[0] /= (float)count;
   n[1] /= (float)count;
   n[2] /= (float)count;
   
   return n;
}

/******************************************************************************/

/*
**  Calculates the normal of 3 vectors.
*/

float *normal( float *p1, float *p2, float *p3 )
{
   float *n;
   
   n = (float *)malloc( 3 * sizeof( float ) );
   
   n[0] = ( ( p2[1] * p3[2] ) - ( p2[2] * p3[1] ) ) + 
          ( ( p3[1] * p1[2] ) - ( p3[2] * p1[1] ) ) +
          ( ( p1[1] * p2[2] ) - ( p1[2] * p2[1] ) );
            
   n[1] = ( ( p2[2] * p3[0] ) - ( p2[0] * p3[2] ) ) + 
          ( ( p3[2] * p1[0] ) - ( p3[0] * p1[2] ) ) +
          ( ( p1[2] * p2[0] ) - ( p1[0] * p2[2] ) );
            
   n[2] = ( ( p2[0] * p3[1] ) - ( p2[1] * p3[0] ) ) + 
          ( ( p3[0] * p1[1] ) - ( p3[1] * p1[0] ) ) +
          ( ( p1[0] * p2[1] ) - ( p1[1] * p2[0] ) );
            
   return n;
}