/*
**  group.c
**
**  Shape From Motion utility.
**
**  
**  Shape From Motion is used to derive 3D information about an object by
**  analyzing images of its rotation around a vertical axis.
**
**  This utility takes as input the .TGA frames specified in the command line
**  and provides as output a .RAW file containing the 3D triangular surfaces
**  it derives from the frames.  This file can be imported directly into the
**  Persistence of Vision Raytracer or can be converted to .DXF format for use
**  in popular CAD applications using pre-existing conversion utilities.
**
**  This software may be redistributed freely as long as any modifications of
**  the code are well documented and include the following copyright.
**
**  Copyright (C) 1995, Benjamin Nason Lipchak
*/

#include "group.h"

/******************************************************************************/

/*
**  This module takes the rectangular coordinates 
**  and groups them into triangular surfaces.
*/

int group( float *x, float *y, float *z, int y_q, int a_q, 
           int p, int t_p, int r, int v, char *out )
{
   int num_triangles = 2 * y_q * a_q;
   int lon, lat, degen = 0;
   float *triangles_1;
   float *triangles_2;
   float *triangles_3;
   char t1x[9], t1y[9], t1z[9], t2x[9], t2y[9], t2z[9], t3x[9], t3y[9], t3z[9];
      
/*
**  The coordinates must be grouped into triangles before being exported in
**  .pov or .raw format.  The top and bottom are already made of triangles.
**  The rest of the object is made of quadralaterals --  2 triangles.
*/
  
   if ( ( triangles_1 = (float *)malloc( sizeof(float) * 3 * num_triangles ) )
        == NULL )
   {
      printf( "Error allocating memory!\n" );
      return -1;
   }       
   
   if ( ( triangles_2 = (float *)malloc( sizeof(float) * 3 * num_triangles ) )
        == NULL )
   {
      printf( "Error allocating memory!\n" );
      free( triangles_1 );
      return -1;
   }       
   if ( ( triangles_3 = (float *)malloc( sizeof(float) * 3 * num_triangles ) )
        == NULL )
   {
      printf( "Error allocating memory!\n" );
      free( triangles_1 );
      free( triangles_2 );
      return -1;
   }       

/*
**  First deal with polar regions.
*/

   if ( v == ON )
      printf( "Creating triangular surfaces... \n" );

/*
**  North polar circle.
*/

   for ( lon = 0; lon < a_q; lon++ )
   {
      triangles_1[(lon - degen) * 3] = x[0];
      triangles_1[(lon - degen) * 3 + 1] = y[0];
      triangles_1[(lon - degen) * 3 + 2] = z[0];

      triangles_2[(lon - degen) * 3]     = x[lon + 2];
      triangles_2[(lon - degen) * 3 + 1] = y[lon + 2];
      triangles_2[(lon - degen) * 3 + 2] = z[lon + 2];

      triangles_3[(lon - degen) * 3]   = x[(lon + 1) % a_q + 2];
      triangles_3[(lon - degen) * 3+1] = y[(lon + 1) % a_q + 2];
      triangles_3[(lon - degen) * 3+2] = z[(lon + 1) % a_q + 2];

/*
**  Checking for degenerate triangle (colinear vertices)
*/
   
      sprintf( t1x, "%#.3f", triangles_1[(lon - degen) * 3] );
      sprintf( t1y, "%#.3f", triangles_1[(lon - degen) * 3 + 1] );
      sprintf( t1z, "%#.3f", triangles_1[(lon - degen) * 3 + 2] );
      sprintf( t2x, "%#.3f", triangles_2[(lon - degen) * 3] );
      sprintf( t2y, "%#.3f", triangles_2[(lon - degen) * 3 + 1] );
      sprintf( t2z, "%#.3f", triangles_2[(lon - degen) * 3 + 2] );
      sprintf( t3x, "%#.3f", triangles_3[(lon - degen) * 3] );
      sprintf( t3y, "%#.3f", triangles_3[(lon - degen) * 3 + 1] );
      sprintf( t3z, "%#.3f", triangles_3[(lon - degen) * 3 + 2] );
      
      if ( colinear( t1x, t1y, t1z, t2x, t2y, t2z, t3x, t3y, t3z ) )
         degen++;
   }

/*
**  South polar circle.
*/

   for ( lon = 0; lon < a_q; lon++ )
   {
      triangles_1[((lon + a_q) - degen) * 3] = x[1];
      triangles_1[((lon + a_q) - degen) * 3 + 1] = y[1];
      triangles_1[((lon + a_q) - degen) * 3 + 2] = z[1];
      
      triangles_2[((lon+a_q)-degen)*3]   = x[((y_q - 1) * a_q) + lon + 2];
      triangles_2[((lon+a_q)-degen)*3+1] = y[((y_q - 1) * a_q) + lon + 2];
      triangles_2[((lon+a_q)-degen)*3+2] = z[((y_q - 1) * a_q) + lon + 2];
      
      triangles_3[((lon+a_q)-degen)*3]  =
                             x[((y_q - 1) * a_q) + (lon + 1) % a_q + 2];
      triangles_3[((lon+a_q)-degen)*3+1]=
                             y[((y_q - 1) * a_q) + (lon + 1) % a_q + 2];
      triangles_3[((lon+a_q)-degen)*3+2]=
                             z[((y_q - 1) * a_q) + (lon + 1) % a_q + 2];

      sprintf( t1x, "%#.3f", triangles_1[((lon+a_q) - degen) * 3] );
      sprintf( t1y, "%#.3f", triangles_1[((lon+a_q) - degen) * 3 + 1] );
      sprintf( t1z, "%#.3f", triangles_1[((lon+a_q) - degen) * 3 + 2] );
      sprintf( t2x, "%#.3f", triangles_2[((lon+a_q) - degen) * 3] );
      sprintf( t2y, "%#.3f", triangles_2[((lon+a_q) - degen) * 3 + 1] );
      sprintf( t2z, "%#.3f", triangles_2[((lon+a_q) - degen) * 3 + 2] );
      sprintf( t3x, "%#.3f", triangles_3[((lon+a_q) - degen) * 3] );
      sprintf( t3y, "%#.3f", triangles_3[((lon+a_q) - degen) * 3 + 1] );
      sprintf( t3z, "%#.3f", triangles_3[((lon+a_q) - degen) * 3 + 2] );
      
      if ( colinear( t1x, t1y, t1z, t2x, t2y, t2z, t3x, t3y, t3z ) )
         degen++;
   }      

/*
**  The rest of the beast...
*/

   for ( lat = 0; lat < y_q - 1; lat++ )
      for ( lon = 0; lon < a_q; lon++ )
      {
         
/*
**  Each quad is two triangles... 1st one:
*/      

         triangles_1[((a_q * 2)+(lat * 2 * a_q)+(lon * 2)-degen)*3] =
            x[2 + (lat * a_q) + lon]; 
         triangles_1[((a_q * 2)+(lat * 2 * a_q)+(lon * 2)-degen)*3+1] =
            y[2 + (lat * a_q) + lon];
         triangles_1[((a_q * 2)+(lat * 2 * a_q)+(lon * 2)-degen)*3+2] =
            z[2 + (lat * a_q) + lon];

         triangles_2[((a_q * 2)+(lat * 2 * a_q)+(lon * 2)-degen)*3] =
            x[2 + (lat * a_q) + (lon + 1) % a_q];
         triangles_2[((a_q * 2)+(lat * 2 * a_q)+(lon * 2)-degen)*3+1] =
            y[2 + (lat * a_q) + (lon + 1) % a_q];
         triangles_2[((a_q * 2)+(lat * 2 * a_q)+(lon * 2)-degen)*3+2] =
            z[2 + (lat * a_q) + (lon + 1) % a_q];
         
         triangles_3[((a_q * 2)+(lat * 2 * a_q)+(lon * 2)-degen)*3] =
            x[2 + (lat + 1) * a_q + lon];
         triangles_3[((a_q * 2)+(lat * 2 * a_q)+(lon * 2)-degen)*3+1] = 
            y[2 + (lat + 1) * a_q + lon];
         triangles_3[((a_q * 2)+(lat * 2 * a_q)+(lon * 2)-degen)*3+2] =
            z[2 + (lat + 1) * a_q + lon];

         sprintf( t1x, "%#.3f", 
            triangles_1[((a_q*2)+(lat*2*a_q)+(lon*2)-degen) * 3] );
         sprintf( t1y, "%#.3f", 
            triangles_1[((a_q*2)+(lat*2*a_q)+(lon*2)-degen) * 3 + 1] );
         sprintf( t1z, "%#.3f", 
            triangles_1[((a_q*2)+(lat*2*a_q)+(lon*2)-degen) * 3 + 2] );
         sprintf( t2x, "%#.3f", 
            triangles_2[((a_q*2)+(lat*2*a_q)+(lon*2)-degen) * 3] );
         sprintf( t2y, "%#.3f", 
            triangles_2[((a_q*2)+(lat*2*a_q)+(lon*2)-degen) * 3 + 1] );
         sprintf( t2z, "%#.3f", 
            triangles_2[((a_q*2)+(lat*2*a_q)+(lon*2)-degen) * 3 + 2] );
         sprintf( t3x, "%#.3f", 
            triangles_3[((a_q*2)+(lat*2*a_q)+(lon*2)-degen) * 3] );
         sprintf( t3y, "%#.3f", 
            triangles_3[((a_q*2)+(lat*2*a_q)+(lon*2)-degen) * 3 + 1] );
         sprintf( t3z, "%#.3f", 
            triangles_3[((a_q*2)+(lat*2*a_q)+(lon*2)-degen) * 3 + 2] );
      
         if ( colinear( t1x, t1y, t1z, t2x, t2y, t2z, t3x, t3y, t3z ) )
            degen++;

/*
**  2nd one:
*/      

         triangles_1[((a_q * 2)+(lat * 2 * a_q)+(lon * 2)+1-degen)*3] =
            x[2 + (lat + 1) * a_q + (lon + 1) % a_q]; 
         triangles_1[((a_q * 2)+(lat * 2 * a_q)+(lon*2)+1-degen)*3+1] =
            y[2 + (lat + 1) * a_q + (lon + 1) % a_q];
         triangles_1[((a_q * 2)+(lat * 2 * a_q)+(lon*2)+1-degen)*3+2] =
            z[2 + (lat + 1) * a_q + (lon + 1) % a_q];

         triangles_3[((a_q * 2)+(lat * 2 * a_q)+(lon * 2)+1-degen)*3] =
            x[2 + (lat * a_q) + (lon + 1) % a_q];
         triangles_3[((a_q * 2)+(lat * 2 * a_q)+(lon*2)+1-degen)*3+1] =
            y[2 + (lat * a_q) + (lon + 1) % a_q];
         triangles_3[((a_q * 2)+(lat * 2 * a_q)+(lon*2)+1-degen)*3+2] =
            z[2 + (lat * a_q) + (lon + 1) % a_q];
         
         triangles_2[((a_q * 2)+(lat * 2 * a_q)+(lon * 2)+1-degen)*3] =
            x[2 + (lat + 1) * a_q + lon];
         triangles_2[((a_q * 2)+(lat * 2 * a_q)+(lon*2)+1-degen)*3+1] = 
            y[2 + (lat + 1) * a_q + lon];
         triangles_2[((a_q * 2)+(lat * 2 * a_q)+(lon*2)+1-degen)*3+2] =
            z[2 + (lat + 1) * a_q + lon];

         sprintf( t1x, "%#.3f", 
            triangles_1[((a_q*2)+(lat*2*a_q)+(lon*2)+1-degen) * 3] );
         sprintf( t1y, "%#.3f", 
            triangles_1[((a_q*2)+(lat*2*a_q)+(lon*2)+1-degen) * 3 + 1] );
         sprintf( t1z, "%#.3f", 
            triangles_1[((a_q*2)+(lat*2*a_q)+(lon*2)+1-degen) * 3 + 2] );
         sprintf( t2x, "%#.3f", 
            triangles_2[((a_q*2)+(lat*2*a_q)+(lon*2)+1-degen) * 3] );
         sprintf( t2y, "%#.3f", 
            triangles_2[((a_q*2)+(lat*2*a_q)+(lon*2)+1-degen) * 3 + 1] );
         sprintf( t2z, "%#.3f", 
            triangles_2[((a_q*2)+(lat*2*a_q)+(lon*2)+1-degen) * 3 + 2] );
         sprintf( t3x, "%#.3f", 
            triangles_3[((a_q*2)+(lat*2*a_q)+(lon*2)+1-degen) * 3] );
         sprintf( t3y, "%#.3f", 
            triangles_3[((a_q*2)+(lat*2*a_q)+(lon*2)+1-degen) * 3 + 1] );
         sprintf( t3z, "%#.3f", 
            triangles_3[((a_q*2)+(lat*2*a_q)+(lon*2)+1-degen) * 3 + 2] );
      
         if ( colinear( t1x, t1y, t1z, t2x, t2y, t2z, t3x, t3y, t3z ) )
            degen++;
      }      
      
   if ( v == ON )
      printf( "Removed %d degenerate triangles!\n", degen );         
   num_triangles -= degen;

/*
**  Send triangles to be exported to a usable format.
*/

   if ( export( triangles_1, triangles_2, triangles_3, num_triangles, 
                p, t_p, r, v, out ) )
   {
      printf( "Error exporting surfaces!\n" );
      free( triangles_1 );
      free( triangles_2 );
      free( triangles_3 );
      return -1;
   }

/*
**  Free dynamically allocated memory.
*/

   free( triangles_1 );
   free( triangles_2 );
   free( triangles_3 );

/*
**  Exit gracefully.
*/   
   
   return 0;
}

/******************************************************************************/

/*
**  This routine checks for degenerate triangles (those with colinear vertices).
*/

int colinear( char *sv1x, char *sv1y, char *sv1z,
              char *sv2x, char *sv2y, char *sv2z,
              char *sv3x, char *sv3y, char *sv3z )
{
   int badx = 0, bady = 0, badz = 0;
   
   float v1x, v1y, v1z, v2x, v2y, v2z, v3x, v3y, v3z;
   float d1x, d1y, d1z, d2x, d2y, d2z;
   
   return 0;
   
   v1x = atof( sv1x ); v1y = atof( sv1y ); v1z = atof( sv1z ); 
   v2x = atof( sv2x ); v2y = atof( sv2y ); v2z = atof( sv2z ); 
   v3x = atof( sv3x ); v3y = atof( sv3y ); v3z = atof( sv3z ); 

   d1x = v1x - v2x; d1y = v1y - v2y; d1z = v1z - v2z;
   d2x = v1x - v3x; d2y = v1y - v3y; d2z = v1z - v3z;

   if ( d1x == 0 || d2x == 0 )
      badx = 1;
   if ( d1y == 0 || d2y == 0 )
      bady = 1;
   if ( d1z == 0 || d2z == 0 )
      badz = 1;
      
   if ( !bady && !badz )
   {
      if ( (d1x/d1y) == (d2x/d2y) && (d1x/d1z) == (d2x/d2z) )
         return 1;
      else
         return 0;
   }  else if ( !bady && !badx )
   {
      if ( (d1z/d1x) == (d2z/d2x) && (d1z/d1y) == (d2z/d2y) )
         return 1;
      else
         return 0;
   }  else if ( !badx && !badz )
   {
      if ( (d1y/d1x) == (d2y/d2x) && (d1y/d1z) == (d2y/d2z) )
         return 1;
      else
         return 0;
   }  else
      return 1;
   
}