/*
**  merge.c
**
**  Shape From Motion utility.
**
**  
**  Shape From Motion is used to derive 3D information about an object by
**  analyzing images of its rotation around a vertical axis.
**
**  This utility takes as input the .TGA frames specified in the command line
**  and provides as output a .POV file containing the 3D triangular surfaces
**  it derives from the frames.  This file can be imported directly into the
**  Persistence of Vision Raytracer or can be converted to .DXF format for use
**  in popular CAD applications using pre-existing conversion utilities.
**
**  This software may be redistributed freely as long as any modifications of
**  the code are well documented and include the following copyright.
**
**  Copyright (C) 1995, Benjamin Nason Lipchak
*/

#include "merge.h"

/******************************************************************************/

/*
**  Here is where the quantization and surface generation takes place...
**
**  Right now it is just an array with triplets containing the angle, the 
**  height, and the radius in cylindrical coordinates.
**
**  For each integral height and quantized angle, a radius will be drawn
**  from the median of all available radii for that height and angle.
**
**  Then the radii will be grouped into triangular surfaces and exported
**  to either a POV scene file and/or a RAW triangle file.
*/

int merge( double *radii, long int total, char *out, int height, int p, 
           int t_p, int r, int v, int a, int c, int q_i, int q_s, int q_e )
{
   int flag, total_vals, bottom, top;
   int line_incomplete;
   int quantum, ty, ta;
   long int i;
   double y, angle;
   double *new_radii, *quantized;

/*
**  Try at highest set resolution.  If not enough info, degrade by 1 degree.
**  Exit with error if degraded below lowest set resolution.
*/
   
   for ( quantum = q_s, flag = 0; 
         quantum >= q_e && flag == 0; 
         quantum -= q_i )
   {
      if ( v == ON )
      {
         printf( "\n\n\nQuantizing to the nearest %g degrees and %.2g height units... ",
                 (360.0 / (double)quantum), ((double)height / (double)(quantum + 1)) );
         fflush( stdout );
      }
 
      if ( ( new_radii = (double *)malloc( sizeof( double ) * quantum * 
                                           quantum ) ) == NULL )
      {
         printf( "\n\nUnable to allocate memory for new_radii!\n\n" );
         return -1;
      }
        
/*
**  Loop through heights and angles
*/

      for ( ty = 0, bottom = -1, top = -1; ty < quantum; ty++ )
      {
         if ( ty == 0 )
            printf( "\n\t" );
         else
            printf( "\b\b\b\b\b\b\b\b\b\b\b\b" );
         
         printf( "%2d%% done... ", ty * 100 / quantum );
         fflush( stdout );

         y = (double)(ty + 1) * ((double)height / (double)(quantum + 1));
         
         for ( ta = 0, line_incomplete = 0; ta < quantum; ta++ )
         {
            angle = (double)ta * (2.0 * PI / (double)quantum);
            
            if ( ( quantized = 
                        (double *)malloc( sizeof( double ) * 300 ) ) == NULL )
            {
               printf( "\n\nUnable to allocate tempory array!\n\n" );
               free( new_radii );
               return -1;
            }

/*
**  Check for radii that match height and quantized angle.
*/
            
            for ( i = 0, total_vals = 0; i < total; i++ )
            {
               double gty = y - ((double)height / (double)(quantum + 1));
               double lty = y + ((double)height / (double)(quantum + 1));

               double gta = fmod( ( (2.0*PI) + angle - (2.0*PI/(double)quantum) ), 
                                  (2.0 * PI) );
               double lta = angle + (2.0 * PI / (double)quantum);
               
               if ( angle == 0.0 && 
                    radii[i*3 + 1] > gty && radii[i*3 + 1] < lty &&
                    ( radii[i*3] > gta || radii[i*3] < lta ) )
               {
                  quantized[total_vals] = radii[i*3 + 2];
                  total_vals++;
                  if ( total_vals%300 == 0 )
                     if ( ( quantized = (double *)realloc( quantized,
                            sizeof( double ) * ( total_vals + 300 ) ) ) == NULL )
                     {
                        printf( "\n\nUnable to reallocate tempory array!\n\n" );
                        free( new_radii );
                        return -1;
                     }
               }  else if ( radii[i*3 + 1] > gty && radii[i*3 + 1] < lty && 
                            radii[i*3] > gta && radii[i*3] < lta )
               {
                  quantized[total_vals] = radii[i*3 + 2];
                  total_vals++;
                  if ( total_vals%300 == 0 )
                     if ( ( quantized = (double *)realloc( quantized,
                            sizeof( double ) * ( total_vals + 300 ) ) ) == NULL )
                     {
                        printf( "\n\nUnable to reallocate tempory array!\n\n" );
                        free( new_radii );
                        return -1;
                     }
               }
            }
               
            if ( total_vals == 0 )
            {
               line_incomplete = 1;
               ta = quantum;
            }   
            else

/*
**  Assign median value to new_radii.
*/
               new_radii[ty*quantum + ta] = median( quantized, total_vals );   
            
            free( quantized );
         }
/*
**  Check for top, bottom, and gaps.
*/
            
         if ( line_incomplete == 1 && top != -1 && bottom == -1 )
            bottom = ty - 1;
         else if ( line_incomplete == 0 && top == -1 )
            top = ty;
         else if ( line_incomplete == 0 && bottom != -1 )  
         {
            if ( v == ON )
               printf( "No good @%g!\n\n", y );
            ty = quantum;
            free( new_radii );
         }
      }
/*
**  Check to see if we're here because we're successful.
*/         

      if ( ty == quantum && ( bottom > top || bottom == -1 ) )
      {
         flag = 1;
         if ( bottom == -1 )
            bottom = quantum - 1;
      }
   }
   
   if ( flag == 0 )
   {
      printf( "\nRadius map is incomplete for all quanta.  Try adjusting parameters.\n\n" );
      return -1;
   }   

   if ( v == ON && top != -1 )
      printf( "\aSuccessful!\n\n" );
   else
   {
      printf( "\nNo object in image!\n\n" );
      free( new_radii );
      return -1;
   }
   
/*
**  Send radii to be converted to rectangular coordinates.
*/

   if ( coords( out, new_radii, quantum + q_i, top, bottom, height, 
                p, t_p, r, a, v ) )
   {
      printf( "An error ocurred in the output module!\n" );
      free( new_radii );
      return -1;
   }   
      
/*
**  Free dynamically allocated memory.
*/

   free( new_radii );      
      
/*
**  Exit gracefully.
*/
 
   return 0;
}              

/******************************************************************************/

/*
**  Returns median of total number of double precision numbers
*/

double median( double *values, int total )
{
   int i, middle_pos;
   double med;
   NODE *first, *current, *comp;
   
   middle_pos = total / 2;
   if ( total%2 == 1 )
      middle_pos++;
   
   for ( i = 0; i < total; i++ )
   {
      if ( ( current = (NODE *)malloc( sizeof( NODE ) ) ) == NULL )
      {
         printf( "Cannot allocate memory for median routine!\n" );
         return -1.0;
      }
      
      current->val = values[i];
      
      if ( i == 0 )
      {
         current->next = NULL;
         first = current;
      }
      else
      {
         if ( current->val < first->val )
         {
            current->next = first;
            first = current;
         } else
         {
            comp = first;
            if ( comp->next == NULL )
            {
               current->next = NULL;
               comp->next = current;
            }  else
            {
               while( current->val > (comp->next)->val )
               {
                  if ( (comp->next)->next != NULL )
                     comp = comp->next;
                  else
                  {
                     comp = comp->next;
                     break;
                  }
               }
               current->next = comp->next;
               comp->next = current;
            }
         }  
      }
   }
   
   for ( i = 0, current = first; i < middle_pos - 1; i++ )
      current = current->next;
   
   med = current->val;
   
   if ( total%2 == 0 )
   {
      current = current->next;
      med += current->val;
      med *= .5;
   }

   for ( i = 0, current = first; i < middle_pos - 1; i++ )
   {
      comp = current;   
      current = current->next;
      free( comp );
   }
   return med; 
}