/*
**  parse.c
**
**  Shape From Motion utility.
**
**  
**  Shape From Motion is used to derive 3D information about an object by
**  analyzing images of its rotation around a vertical axis.
**
**  This utility takes as input the .TGA frames specified in the command line
**  and provides as output a .POV file containing the 3D triangular surfaces
**  it derives from the frames.  This file can be imported directly into the
**  Persistence of Vision Raytracer or can be converted to .DXF format for use
**  in popular CAD applications using pre-existing conversion utilities.
**
**  This software may be redistributed freely as long as any modifications of
**  the code are well documented and include the following copyright.
**
**  Copyright (C) 1995, Benjamin Nason Lipchak
*/

#include "parse.h"

/******************************************************************************/

/*
**  Main module
**
**  Parses the command line.  A base string for the input frames is required.
**  Other variables are optional, with defaults indicated.
*/

int main( int argc, char **argv )
{

/*
**  Variables with their defaults
*/

   int num_frames = LATER;
   int matrix_size = LATER;
   int verbose = ON;
   int center = LATER;
   int raw = OFF;
   int pov = ON;
   int t_pov = OFF;
   int axis = Y;
   int direction = CCW;
   int quantum_inc = 12;
   int quantum_start = 72;
   int quantum_end = 24;

   double sigma = 1.75;
   double best_threshold = .75;
   
   char *output;
   
/*
**  Other variables
*/

   int i;
   int base_length;
   char *base_string;

   output = (char *)malloc( 7 );
   strcpy( output, "output" );
   
/*
**  Begin to parse command line.  Make sure at least base_string is present.
*/
   
   if ( argc < 2 )
   {
      printf( "Usage: tga2pov [options] base_string\n" );
      free( output );
      return 0;
   }

/*
**  Loop through and parse options.
*/

   for ( i = 1; i < argc - 1; i++ )
      if ( strlen( argv[i] ) < 2 || ( argv[i][0] != '-' && argv[i][0] != '+' ) )
      {
         clhelp();
         free( output );
         return 0;
      }  else
      {
         switch( argv[i][1] )
         {
            case 'a':  axis = X;
                       break;
                       
            case 'b':  if ( atof( &argv[i][2] ) > 0.0 )
                          best_threshold = atof( &argv[i][2] );
                       break;
            
            case 'c':  if ( atoi( &argv[i][2] ) > 0 )
                          center = atoi( &argv[i][2] );
                       break;
            
            case 'd':  direction = CW;
                       break;
            
            case 'h':  clhelp();
                       return 0;
            
            case 'i':  if ( atoi( &argv[i][2] ) > 0 )
                          quantum_inc = atoi( &argv[i][2] );
                       break;
            
            case 'j':  if ( atoi( &argv[i][2] ) > 0 )
                          quantum_start = atoi( &argv[i][2] );
                       break;
            
            case 'k':  if ( atoi( &argv[i][2] ) > 0 )
                          quantum_end = atoi( &argv[i][2] );
                       break;
            
            case 'm':  if ( atoi( &argv[i][2] ) > 0 )
                          matrix_size = atoi( &argv[i][2] );
                       break;

            case 'n':  if ( atoi( &argv[i][2] ) > 0 )
                          num_frames = atoi( &argv[i][2] );
                       break;
            
            case 'o':  if ( strlen( &argv[i][2] ) > 0 )
                       {
                          output = (char *)malloc( strlen( &argv[i][2] ) + 1 );
                          strcpy( output, &argv[i][2] );
                       }
                       break;
            
            case 'p':  pov = OFF;
                       break;
            
            case 'r':  raw = ON;
                       break;
            
            case 's':  if ( atof( &argv[i][2] ) > 0.0 )
                          sigma = atof( &argv[i][2] );
                       break;
            
            case 't':  pov = OFF;
                       t_pov = ON;
                       break;
            
            case 'v':  verbose = OFF;
                       break;
            
            default :  clhelp();
                       return 0;
         }
      }

/*
**  Make sure the first character isn't a hyphen so we know it's
**  the base_string.
*/

   base_length = strlen( argv[argc - 1] );
 
   if ( argv[argc - 1][0] == '-' ) 
      if ( argv[argc - 1][1] == 'h' )
      {
         clhelp();
         free( output );
         return 0;
      }  else
      {
         printf( "Usage: tga2pov [options] base_string\n" );
         printf( "base_string must not begin with a hyphen!\n" );
         free( output );
         return 0;
      }

/*  
**  Options and base_string are passed to
**  rest of program.  If verbose on, show parsed command line.
*/
   
   base_string = (char *)malloc( base_length + 1 );
   strcpy( base_string, argv[argc - 1] );
   
   if ( verbose == ON )
   {
      printf( "\nCurrent options: -o%s -n", output );
      
      if ( num_frames == LATER )
         printf( "? -c" );
      else
         printf( "%d -c", num_frames );

      if ( center == LATER )
         printf( "? -a" );
      else
         printf( "%d -a", center );
         
      if ( axis == Y )
         printf( "Y -d" );
      else
         printf( "X -d" );

      if ( direction == CW )
         printf( "CW -r" );
      else
         printf( "CCW -r" );

      if ( raw == ON )
         printf( "ON -p" );
      else
         printf( "OFF -p" );

      if ( pov == ON )
         printf( "ON -t" );
      else
         printf( "OFF -t" );
         
      if ( t_pov == ON )
         printf( "ON -v" );
      else
         printf( "OFF -v" );

      printf( "ON\n                 -s%g -m", sigma ); 

      if ( matrix_size == LATER )
         printf( "? -b%g -i%d -j%d -k%d\n\n", best_threshold, 
                 quantum_inc, quantum_start, quantum_end );
      else
         printf( "%d -b%g -i%d -j%d -k%d\n\n", matrix_size, best_threshold,
                 quantum_inc, quantum_start, quantum_end );
   }
   
/*
**  Start rest of program
*/   

   if ( defaults( base_string, output, num_frames, center, axis, direction,
                  pov, t_pov, raw, verbose, quantum_inc, quantum_start, quantum_end,
                  sigma, matrix_size, best_threshold ) )
   {
      printf( "UNSUCCESSFUL!\n" );
      free( base_string );
      free( output );
      return -1;
   }

/*
**  Free dynamically allocated memory.
*/
   
   free( base_string );
   free( output );
   
/*
**  Graceful exit.
*/

   return 0;
} 

/******************************************************************************/

/*
**  Help screen for syntax and meaning of options
*/

void clhelp( void )
{
   printf( "\n\ntga2pov [options] base_string\t\t" );
   printf( "Copyright 1994, Benjamin Nason Lipchak\n\n" );
   
   printf( "Options: Form\tDescription\t\t\t\tDefault Value\n\n" );
   
   printf( "\t-nXXX\tNumber of frames\t\t\tDetermined by program\n" );
   printf( "\t-cXXX\tCenter of rotation\t\t\tDetermined by program\n" );
   printf( "\t-a\tAxis of rotation = X\t\t\tY\n" );
   printf( "\t-d\tDirection of rotation = CW\t\tCCW\n" );
   printf( "\t-oFILE\tOutput filename = FILE\t\t\toutput\n\n" );
   
   printf( "\t-sX.X\tSigma value for edge detection\t\t1.75\n" );
   printf( "\t-mXX\tMatrix size for correspondence\t\tDetermined by program\n" );
   printf( "\t-b0.X\tBest-match threshold value 0<b<1\t0.75\n\n" );

   printf( "\t-iXX\tIncrement for quantization\t\t12\n" );
   printf( "\t-jXX\tMaximum value for quantization\t\t72\n" );
   printf( "\t-kXX\tMinimum value for quantization\t\t24\n\n" );

   printf( "\t-p\tSmooth POV scene file output = OFF\tON\n" );
   printf( "\t-t\tFaceted POV scene file output = ON\tOFF\n" );
   printf( "\t-r\tRAW triangle output = ON\t\tOFF\n" );
   printf( "\t-v\tVerbose mode = OFF\t\t\tON\n\n" );
}